/*
 *  NEZplug++ VGM mod SCC Fix
 *
 *
 * Original Source 
 *   VGM YM2608 ch1-6 -> YM2612 ch1-6 command converter
 *     for use to compare 2612 vs 2608
 *   
 *   (C) 2014 James Alan Nguyen
 *   http://www.codingchords.com
 *
 * use at your peril
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#pragma pack(1)

#define DEF_BUF_SIZE (80)

int printUsage(char *execName)
{
   printf(
      "SCC Fix"
      "Usage: %s file.vgm\n",
      execName
   );

   return EXIT_SUCCESS;
}

int run(char *fileName)
{
   char nameCopy[DEF_BUF_SIZE];/*= "converted.vgm";*/
   char *fileBuffer = NULL, *curPos = NULL, *bufEnd = NULL;
   FILE *inFile = NULL, *outFile = NULL;
   size_t fileSize = 0;
   int hedType = 0;
   unsigned long int fsize = 0;
   
   /*Do not allow converting .vgz (TODO gzip integration?)*/
   if (strcmp(".vgz", &(fileName[strlen(fileName)-4])) == 0)
   {
      perror(".vgz not supported\n");
      return EXIT_FAILURE;
   }
   
   strncpy(nameCopy, fileName, DEF_BUF_SIZE);
   strncpy(&(nameCopy[strlen(fileName)-4]), "-fix.vgm", DEF_BUF_SIZE); 
   
   inFile  = fopen(fileName, "rb"),
   outFile = fopen(nameCopy, "wb");
   
   if (inFile == NULL || outFile == NULL)
   {
      perror("File cannot be opened.\n");
      return EXIT_FAILURE;
   }
   
   /*Get file size*/
   fseek(inFile, 0, SEEK_END);
   fileSize = ftell(inFile);
   fseek(inFile, 0, SEEK_SET);

   /*Init file in memory*/
   fileBuffer = calloc(sizeof(char), fileSize);
   if (fileBuffer == NULL)
   {
      perror("Cannot initialize file buffer.\n");
      return EXIT_FAILURE;
   }

   /*Copy contents*/
   /*TODO: ASSUMING VGM 1.70+ HEADER*/
   fread(fileBuffer, 1, fileSize, inFile);

   fileBuffer[0x74] = 0x4c;
   fileBuffer[0x75] = 0x4f;
   fileBuffer[0x76] = 0x1b;

   fileBuffer[0x9c] = 0x4c;
   fileBuffer[0x9d] = 0x4f;
   fileBuffer[0x9e] = 0x1b;

   curPos = &(fileBuffer[0xc0]);
   bufEnd = &(fileBuffer[fileSize-1]);

   /*Exhaustively read each VGM command*/
   while (curPos <= bufEnd)
   {
      size_t ptrInc;

      switch(curPos[0] & 0xFF)
      {
			
      case 0x62: /*1-byte Wait 60Hz shortcut*/
      case 0x63: /*1-byte Wait 50Hz shortcut*/
         ptrInc = 1;
         break;
      
      case 0x61: /*3-byte Wait*/
         ptrInc = 3;
         break;

      case 0x66: /*End of sound data*/
         ptrInc = 1;
         curPos = (bufEnd);
         break;
      
      case 0x67: /*Data Block*/
         printf("DEBUG: Curr addr 0x%X\n", curPos-fileBuffer);
         ptrInc = ((*(int*)(curPos+3))); /*32-bit offset*/
         printf("DEBUG: Offset Data Block 0x%X\n", ptrInc);
         curPos += ptrInc + 0x07;
         printf("DEBUG: New addr 0x%X\n", curPos-fileBuffer);
         ptrInc = 0;
         break;
      
      case 0xA0: /*PSG*/
         ptrInc = 3;
         break;

      case 0xD2: /*SCC*/
         ptrInc = 4;
	     if (curPos[1] == 0x01)
	     {
	         printf("DEBUG: FIX ADDR 0x%X\r", (curPos-fileBuffer));
	         curPos[2] = (curPos[2] & 0x0F);
	     }
	     if (curPos[1] == 0x02)
	     {
	         printf("DEBUG: FIX ADDR 0x%X\r", (curPos-fileBuffer));
	         curPos[2] = (curPos[2] & 0x0F);
	     }
	     if (curPos[1] == 0x03)
	     {
	         printf("DEBUG: FIX ADDR 0x%X\r", (curPos-fileBuffer));
	         curPos[2] = (curPos[2] & 0x00);
	     }

/* scc+ -> scc test
	     if (curPos[1] == 0x04)
	     {
	        printf("DEBUG: FIX ADDR 0x%X\r", (curPos-fileBuffer));
                curPos[1] = 0x00;

                if (curPos[2] & 0x80)
                {
                 curPos[1] = 0x03;
	         curPos[2] = 0x00;
	         curPos[3] = 0x00;
                }
	     }
*/

         break;

      default:
         if (curPos[0] & 0xF0) /*1-byte Wait*/
            ptrInc = 1;
	 else
	    ptrInc = 3;
         break;
      }
      
      /*Increment based on prefix command*/
      curPos += ptrInc;
   }

   /*Write to file*/
   fwrite(fileBuffer, 1, fileSize, outFile);
   fclose(outFile);

   return EXIT_SUCCESS;
}

int main(int argc, char *argv[])
{
   if (argc != 2)
      return printUsage(argv[0]);

   return run(argv[1]);
}